/*
 * file FolderView.java
 * 
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.teamapi.scout.FolderView
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.teamapi.scout;

import java.util.Iterator;
import java.util.List;

import javax.wvcm.PropertyRequestItem.PropertyRequest;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.ViewerFilter;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.part.ViewPart;

import com.ibm.rational.wvcm.stp.StpProvider;
import com.ibm.rational.wvcm.stp.cc.CcRegistryRegion;


/**
 * A CM API Folder hierarchy viewer based on the Eclipse TreeViewer control. The
 * hierarchy is determined by the CHILD_BINDING_LIST property of each folder in
 * the view. Each element of the tree view is a ProxyElement, which wraps a CM
 * API proxy and represents the resource referenced by that proxy. Each element
 * implements IPropertySource, so the properties of the CM API resource can be
 * viewed in the Eclipse property viewer simply by selecting the element in the
 * tree view.
 */
public class FolderView
    extends ViewPart
{
    /**
     * The constructor.
     */
    public FolderView() {}

    /**
     * @see  IWorkbenchPart#createPartControl(Composite)
     *
     * @param  parent  The Composite that is the parent of this View.
     */
    public void createPartControl(Composite parent)
    {
        /*
         * Create a grid layout object so the text and tree viewer layout is
         * under our control.
         */
        GridLayout layout = new GridLayout();

        layout.numColumns = 1;
        layout.verticalSpacing = 2;
        layout.marginWidth = 0;
        layout.marginHeight = 2;
        parent.setLayout(layout);

        /*
         * Create a "label" in which to display information about the selected
         * resource. It's a text field instead of a label so user can copy-paste
         * out of it.
         */
        m_text = new Text(parent, SWT.READ_ONLY | SWT.SINGLE | SWT.BORDER);

        // layout the text field above the tree viewer
        GridData layoutData = new GridData();

        layoutData.grabExcessHorizontalSpace = true;
        layoutData.horizontalAlignment = GridData.FILL;
        m_text.setLayoutData(layoutData);

        // Create the tree viewer as a child of the composite parent
        m_treeViewer = new TreeViewer(parent);
        m_treeViewer.setContentProvider(new FolderContentProvider());
        m_treeViewer.setLabelProvider(new FolderLabelProvider());
        m_treeViewer.setUseHashlookup(true);

        // layout the tree viewer below the text field
        layoutData = new GridData();
        layoutData.grabExcessHorizontalSpace = true;
        layoutData.grabExcessVerticalSpace = true;
        layoutData.horizontalAlignment = GridData.FILL;
        layoutData.verticalAlignment = GridData.FILL;
        m_treeViewer.getControl().setLayoutData(layoutData);

        // Create menu, tool bars, filters, sorters.
        createFiltersAndSorters();
        createActions();
        createMenus();
        createToolbar();
        hookListeners();

        m_treeViewer.setInput(getInitalInput(parent.getShell()));
        getSite().setSelectionProvider(m_treeViewer);
    }

    /**
     * Creates the filtering and sort objects used to filter and sort the
     * display
     */
    protected void createFiltersAndSorters()
    {
        m_atLeastThreeFilter = new NonEmptyFolderFilter();
        m_onlyFoldersFilter = new FolderOnlyFilter();
        m_foldersFirstSorter = new FoldersFirstSorter();
        m_noTypeSorter = new NoTypeSorter();
    }

    /**
     * Creates selection-changed listener, which keeps the label field
     * up-to-date with the selected elements of the tree view.
     */
    protected void hookListeners()
    {
        m_treeViewer.addSelectionChangedListener(new ISelectionChangedListener() {
                public void selectionChanged(SelectionChangedEvent event)
                {
                    // if the selection is empty clear the label
                    if (event.getSelection().isEmpty()) {
                        m_text.setText("");

                        return;
                    }

                    if (event.getSelection() instanceof IStructuredSelection) {
                        IStructuredSelection selection =
                            (IStructuredSelection)event.getSelection();
                        StringBuffer         toShow = new StringBuffer();

                        for (Iterator iterator = selection.iterator();
                                 iterator.hasNext();) {
                            if (toShow.length() > 0)
                                toShow.append(", ");

                            toShow.append(((ProxyElement)iterator.next())
                                              .getSelector());
                        }

                        m_text.setText(toShow.toString());
                    }
                }
            });
    }

    /**
     * Creates the actions used by this view
     */
    protected void createActions()
    {
        m_onlyFoldersAction = new Action("Only Folders") {
                public void run()
                {
                    updateFilter(m_onlyFoldersAction);
                }
            };
        m_onlyFoldersAction.setChecked(false);

        m_nonEmptyFolders = new Action("No Empty Folders") {
                public void run()
                {
                    updateFilter(m_nonEmptyFolders);
                }
            };
        m_nonEmptyFolders.setChecked(false);

        m_foldersFirstAction = new Action("Folders First") {
                public void run()
                {
                    updateSorter(m_foldersFirstAction);
                }
            };
        m_foldersFirstAction.setChecked(false);

        m_noTypeAction = new Action("Ignore Types") {
                public void run()
                {
                    updateSorter(m_noTypeAction);
                }
            };
        m_noTypeAction.setChecked(false);

        m_addRootAction = new Action("Add Root Folder") {
                public void run()
                {
                    addNewRoot();
                }
            };
        m_addRootAction.setToolTipText("Add a new root folder to this view");
        m_addRootAction.setImageDescriptor(ScoutPlugin.getImageDescriptor("addRoot.gif"));

        m_removeAction = new Action("Elide") {
                public void run()
                {
                    removeRoot();
                }
            };
        m_removeAction.setToolTipText("Remove selected root folder(s) from this view");
        m_removeAction.setImageDescriptor(ScoutPlugin.getImageDescriptor("remove.gif"));

        m_refreshAction = new Action("Refresh") {
                public void run()
                {
                    refresh();
                }
            };
        m_refreshAction.setToolTipText("Recompute bound member list of selected folder(s)");
        m_refreshAction.setImageDescriptor(ScoutPlugin.getImageDescriptor("refresh.gif"));
    }

    /**
     * Raises a dialog requesting the selector for a folder resource to be added
     * to the tree view. Unless canceled, by the user, the new folder is added
     * as a root of the display.
     */
    protected void addNewRoot()
    {
        InputDialog dialog =
            new InputDialog(null, "Add Root Folder",
                            "Enter the Location for a New Root folder."
                            + " Use 'views', 'vobs', or 'dbsets' "
                            + " to add repositories known to server",
                            "<domain>.<namespace>:<name>@<repository>", null);

        if (dialog.open() == InputDialog.OK) {
            String root = dialog.getValue(); 
            StpProvider provider = m_root.getProvider();
            
            try {
                if (m_askForServerUrl) {
                    dialog =
                        new InputDialog(null,
                                        "CM Server URL",
                                        "Enter the URL for your CM Server.",
                                        "http://qwin115:9082/TeamWeb/services/Team",
                                        null);

                    if (dialog.open() == InputDialog.OK)
                        provider.setServerUrl(dialog.getValue());
                    
                    m_askForServerUrl = false;
                }
                
                if (root.equals("dbsets")) {
                    if (addRoots(provider.cqProvider().doGetDbSetList(null)))
                        m_treeViewer.refresh(m_root);
                } else if (root.equals("vobs")) {
                    CcRegistryRegion rr =
                        provider.ccProvider().doGetDefaultCcRegistryRegion
                            (new PropertyRequest(CcRegistryRegion.VOB_TAG_LIST));

                    if (addRoots(rr.getVobTagList()))
                        m_treeViewer.refresh(m_root);
                } else if (root.equals("views")) {
                    CcRegistryRegion rr =
                        provider.ccProvider().doGetDefaultCcRegistryRegion
                            (new PropertyRequest(CcRegistryRegion.VIEW_TAG_LIST));

                    if (addRoots(rr.getViewTagList()))
                        m_treeViewer.refresh(m_root);
                } else if (addRoot(root))
                    m_treeViewer.refresh(m_root);
            } catch (Throwable e) {
                MessageDialog.openError(null, 
                                        "Can't Add " + root,
                                        e.getClass().getName() + ": "
                                        + e.getLocalizedMessage());          
                e.printStackTrace();
            }
        }
    }

    /**
     * Removes the selected object(s) from the tree view. Only root folders at
     * the top level may be removed since all other tree view entries are
     * derived objects.
     */
    protected void removeRoot()
    {
        boolean removed = false;

        try {
            IStructuredSelection selection =
                (IStructuredSelection)m_treeViewer.getSelection();

            for (Iterator iterator = selection.iterator(); iterator.hasNext();)
                removed |= m_root.removeChild(iterator.next());
        } catch (Throwable ex) {
            MessageDialog.openError(null, "Remove Root Failed", ex.getClass()
                                        .getName() + ": "
                                    + ex.getLocalizedMessage());
        }

        if (removed)
            m_treeViewer.refresh(m_root);
    }

    /**
     * Refreshes the child set of each selected resource. The bound member list
     * of each selected child is reread from the repository.
     */
    protected void refresh()
    {
        try {
            IStructuredSelection selection =
                (IStructuredSelection)m_treeViewer.getSelection();

            for (Iterator iterator = selection.iterator();
                    iterator.hasNext();) {
                ProxyElement elem = (ProxyElement)iterator.next();

                m_treeViewer.collapseToLevel(elem, TreeViewer.ALL_LEVELS);
                elem.refresh();
                m_treeViewer.expandToLevel(elem, 1);
            }
        } catch (Throwable ex) {
            MessageDialog.openError(null, "Remove Root Failed", ex.getClass()
                                        .getName() + ": "
                                    + ex.getLocalizedMessage());
        }
    }

    /**
     * Creates the menu structures needed for this view.
     */
    protected void createMenus()
    {
        IMenuManager rootMenuManager =
            getViewSite().getActionBars().getMenuManager();

        rootMenuManager.setRemoveAllWhenShown(true);
        rootMenuManager.addMenuListener(new IMenuListener() {
                public void menuAboutToShow(IMenuManager mgr)
                {
                    fillMenu(mgr);
                }
            });

        fillMenu(rootMenuManager);
    }

    /**
     * Builds the drop-down menu used for this tree view.
     *
     * @param  rootMenuManager  The menu manager that will contain the menu
     *                          items.
     */
    protected void fillMenu(IMenuManager rootMenuManager)
    {
        rootMenuManager.add(m_addRootAction);
        rootMenuManager.add(m_removeAction);
        rootMenuManager.add(m_refreshAction);

        IMenuManager filterSubmenu = new MenuManager("Filters");
        rootMenuManager.add(filterSubmenu);
        filterSubmenu.add(m_onlyFoldersAction);
        filterSubmenu.add(m_nonEmptyFolders);

        IMenuManager sortSubmenu = new MenuManager("Sort By");
        rootMenuManager.add(sortSubmenu);
        sortSubmenu.add(m_foldersFirstAction);
        sortSubmenu.add(m_noTypeAction);
    }

    /**
     * Updates the TreeViewer's sorter to match the options selected by the user
     * in the sort sub-menu.
     *
     * @param  action  The Action selected by the user.
     */
    protected void updateSorter(Action action)
    {
        if (action == m_foldersFirstAction) {
            m_noTypeAction.setChecked(!m_foldersFirstAction.isChecked());

            if (action.isChecked()) {
                m_treeViewer.setSorter(m_foldersFirstSorter);
            } else {
                m_treeViewer.setSorter(null);
            }
        } else if (action == m_noTypeAction) {
            m_foldersFirstAction.setChecked(!m_noTypeAction.isChecked());

            if (action.isChecked()) {
                m_treeViewer.setSorter(m_noTypeSorter);
            } else {
                m_treeViewer.setSorter(null);
            }
        }
    }

    /**
     * Updates the TreeViewer filters to reflect the filtering options selected
     * by the user.
     *
     * @param  action  An Action identifying the filter option that has changed.
     */
    protected void updateFilter(Action action)
    {
        if (action == m_nonEmptyFolders) {
            if (action.isChecked()) {
                m_treeViewer.addFilter(m_atLeastThreeFilter);
            } else {
                m_treeViewer.removeFilter(m_atLeastThreeFilter);
            }
        } else if (action == m_onlyFoldersAction) {
            if (action.isChecked()) {
                m_treeViewer.addFilter(m_onlyFoldersFilter);
            } else {
                m_treeViewer.removeFilter(m_onlyFoldersFilter);
            }
        }
    }

    /**
     * Adds buttons to the tool bar for the actions defined for this view.
     */
    protected void createToolbar()
    {
        IToolBarManager toolbarManager =
            getViewSite().getActionBars().getToolBarManager();
        toolbarManager.add(m_addRootAction);
        toolbarManager.add(m_removeAction);
        toolbarManager.add(m_refreshAction);
    }

    /**
     * Constructs a proxy for the resource named by a given selector and adds it
     * to the tree view as a new top-level folder. The resource referenced by
     * the selector is not accessed until the new entry is highlighted or
     * expanded.
     *
     * @param  selectorString  A String containing a CM API object selector
     *                         naming a resource.
     *
     * @return  true if the resource was successfully added to the tree view.
     */
    protected boolean addRoot(String selectorString)
    {
        try {
            m_root.addChild(selectorString);

            return true;
        } catch (Throwable ex) {
            MessageDialog.openError(null, "Can't Add " + selectorString,
                                    ex.getClass().getName() + ": "
                                    + ex.getLocalizedMessage());

            return false;
        }
    }

    /**
     * Adds the objects in a given list to the display.
     *
     * @param  roots  An array of object whose toString() value is a valid
     *                object selector
     *
     * @return  true if at least one root was added to the display.
     */
    protected boolean addRoots(List roots)
    {
        Iterator iter = roots.iterator();
        boolean  added = false;

        while (iter.hasNext())
            added |= addRoot(iter.next().toString());

        return added;
    }

    /**
     * Makes initial entries into the tree viewer for the sample databases and
     * projects usually installed with ClearQuest. Using a known server (which
     * will have to be changed for your location), a ClearCase server is
     * contacted and client workspaces known to the server are added as roots.
     * 
     * @param shell The Shell used for display context when requesting
     *            credentials from the user.
     * 
     * @return A ProxyElement object representing the (invisible) root of the
     *         tree to be viewed.
     */
    public Object getInitalInput(Shell shell)
    {
        try {
            m_root = new ProxyElement(shell);
            StpProvider provider = m_root.getProvider();

            String url = System.getProperty("scout.cc.server");
            
            if (url != null) {
                provider.setServerUrl(url);
                m_askForServerUrl = false;
            }

            return m_root;
        } catch (Throwable ex) {
            ex.printStackTrace();

            return m_root = null;
        }
    }

    /**
     * @see  IWorkbenchPart#setFocus()
     */
    public void setFocus() {}

    /** The TreeViewer control on which this view is based */
    protected TreeViewer m_treeViewer;

    /** The Text for this view */
    protected Text m_text;

    /** The Actions for implementing display options */
    protected Action m_onlyFoldersAction, m_nonEmptyFolders;

    /** The Actions for implementing sort options */
    protected Action m_foldersFirstAction, m_noTypeAction;

    /** The Actions for implementing list management operations */
    protected Action m_addRootAction, m_removeAction, m_refreshAction;

    /** The ViewFilter for implementing display options */
    protected ViewerFilter m_onlyFoldersFilter, m_atLeastThreeFilter;

    /** The ViewSorter for implementing sort options */
    protected ViewerSorter m_foldersFirstSorter, m_noTypeSorter;

    /** The root of the tree view. */
    protected ProxyElement m_root;
    
    private boolean m_askForServerUrl = true;
}
